/*
 ***********************************************************************
 *
 *					 			Set of 3D Points
 *
 * The present file defines classes to handle a set of 3D points, i.e.
 * to display them using the appropritae projection and perform some
 * linear space transformation (rotation, translation, etc).
 *
 ***********************************************************************
 */

#include "SpacePoints.h"
#include "Database.h"
#include "a_menu.h"

/*
 *----------------------------------------------------------------------
 *						Construct the set of SpacePoints
 */
 
 										// Dummy constructor
SpacePoints::SpacePoints(void)
{
	npoints = 0;
	points = (void *)0;
	picture = (void *)0;
	S.x = 400; S.y = 50; S.z = 550;
}

										// Real destructor
SpacePoints::~SpacePoints(void)
{
	assert( npoints > 0 && points != (void *)0 );
	DisposePtr(points);
}

										// A constructor that really does smth
void SpacePoints::allocate(const short _npoints)
{
	npoints = _npoints;
	assert( npoints > 0 );
	points = (SpacePoint *)NewPtr(npoints*sizeof(SpacePoint));
	assert( points != (void *)0 );
	proj_points = (Point *)NewPtr(npoints*sizeof(Point));
	assert( proj_points != (void *)0 );
}

/*
 *----------------------------------------------------------------------
 *				Some elementary operations on the SpacePoints
 */

										// Set the value of a SpacePoint
										// Note the points are numbered starting from 1
void SpacePoints::set(const short point_no, const short x, const short y, const short z)
{
	assert( npoints > 0 );
	assert( point_no > 0 && point_no <= npoints );
	points[point_no-1].x = x;
	points[point_no-1].y = y;
	points[point_no-1].z = z;
}

void SpacePoints::set_proj_rect
	(const short left, const short top, const short right, const short bottom)
{
	SetRect(&proj_rect,left,top,right,bottom);
}

/*
 *----------------------------------------------------------------------
 *				Project SpacePoints to the plane to display them later
 * Orthogonal projection of the spacepoint is a linear transformation
 *
 *	|h|		| a11 a12 a13 | |x|
 *	|v|	= 	| a21 a22 a23 | |y|
 *  |0|		| 0   0   0	  | |z|
 *
 * For a dimetric projection, the matrix is to satisfy conditions:
 *		(1,0,0)' -> (-a11,-a21,0)' such that a11^2 + a21^2 = 0.25 and a11 = a21
 *		(0,1,0)' -> (1,0,0)'
 *		(0,0,1)  -> (0,1,0)'
 * In other words, axis Oz maps to Ox', Oy is projected to Ox', and vector i maps to
 * a vector of half size pointing away from Oy' by 135 deg.
 * i.e the matrix is
 *						| -11/32  1	 0 |
 *						| -11/32  0  1 |
 *						|    0    0  0 |
 *
 * For the isometric projection, the conditions are
 *		(1,0,0)' -> (a11,a21,0)'
 *		(0,1,0)' -> (a12,a22,0)'
 *		(0,0,1)  -> (0,1,0)'
 * a11^2 + a21^2 = a12^2 + a22^2 = 1 (isometry along three principal directions)
 * cos( Oz',Ox' ) = a21 = cos( 2/3 ) = -sqrt(3)/2
 * cos( Oz',Oy' ) = a22 = cos( 2/3 ) = -sqrt(3)/2
 * i.e the matrix is
 *						|  -1/2     1/2	   0 |
 *						| -27/32  -27/32   1 |
 *						|    0        0    0 |
 *
 * Prospective projection is somewhat different.
 * With point S, (sx,sy,sz), as a projection center and plane zOy as a projection plane,
 * every point A, (x,y,z), gets projected onto point A' of the projection plane such that
 * three points, S, A, and A' lie on the same line. It means that SA' = SA, and
 * point A' has coordinates ((x-sx)+sx, (y-sy)+sy, (z-sz)+sz). Since A' lies on the
 * projection plane, its abscissa should be zero, which gives us the condition to
 * determine scale factor 
 *				(x-sx)+sx = 0
 * or
 *				 = -sx/(x-sx)
 *
 * Prospective projection is not defined for the points with x = sx. We just project them
 * to "nowhere", (-1000,-1000)
 */

void SpacePoints::project(void)
{
	register int i;
	register SpacePoint * sp = points;
	register Point *      pp = proj_points;
	assert( npoints > 0 );
	
	for(; sp < points + npoints; sp++, pp++)
	   switch(Projection_mode)
	   {
		case DIMETRIC:
			 pp->h = (-11*sp->x + 32*sp->y)/32;
			 pp->v = (-11*sp->x + 32*sp->z)/32;
			 break;
			 
		case ISOMETRIC:
			 pp->h = ( -(sp->x) + (sp->y) )/2;
			 pp->v = ( -27*(sp->x) - 27*(sp->y) + 32*(sp->z) )/32;
			 break;
			 
		case PROSPECTIVE:
			 if( S.x == sp->x )
			   pp->h = pp->v = -1000;
			 else
			 {
			   register int m64 = (-64*S.x)/(sp->x - S.x);	// Measured in 64th of a unit	
			   pp->h = (m64*(sp->y - S.y))/64 + S.y;
			   pp->v = (m64*(sp->z - S.z))/64 + S.z;
			 }
			 break;
			 
		default:
			 assure(0,"May not happen");	 
	   }
	
}

/*
 *----------------------------------------------------------------------
 *								Displaying
 */

void SpacePoints::refresh(void)
{
	project();
	if( picture != (void *)0 )				// Make draw create a new picture
	  KillPicture(picture);
	picture = (void *)0;
	inherited::refresh();
}


void SpacePoints::draw_picture(void)
{
	_error("Derived classes should have done smth here");
}

											// Note, at this point grafport is set
											// to our window
void SpacePoints::draw(void)
{
	Rect rect = thePort->portRect;
	Rect clip;

	if( picture == (void *)0 )				// Create a new picture if necessary
	{

										// Note, the clip rectangle should be specified
										// in relative (relative to portBits.bounds
										// rectangle)
		SetRect(&clip,0,0,rect.right-rect.left,rect.bottom-rect.top);
		ClipRect(&clip);				// This is necessary for picture processing 
		picture = OpenPicture(&rect);
		draw_picture();
		ClosePicture();
	}
	BackColor(whiteColor);

	DrawPicture(picture,&thePort->portRect);
}

/*
 *----------------------------------------------------------------------
 *						Picture drawing functions
 * 
 * The functions below are to be used in drawing the Picture. It means Picture
 * mode should be on (set by draw()). They functions must be called by draw() or 
 * its descendants
 */

									// Connect point1 with point2, etc.
									// Terminate the list with 0
void SpacePoints::line_segments(const short point1, const short point2,...)
{
	long * args;
	register int extra_pt_no;
	Rect port_rect = thePort->portRect;
									// The following constants are measured in 8th
									// of the original scale
	const int o_scale_x = (8*(port_rect.right-port_rect.left))/(proj_rect.right-proj_rect.left);
	const int o_scale_y = (8*(port_rect.top-port_rect.bottom))/(proj_rect.top-proj_rect.bottom);

	args = (long *)(&point2) + 1;					// va_start(args,point2);  Init 'args' 
	assert( thePort->picSave != (void *)0 );		// Picture should be under conctruction

	assert( point1 > 0 && point1 <= npoints );
	MoveTo((proj_points[point1-1].h-proj_rect.left)*o_scale_x/8,
		   (proj_points[point1-1].v-proj_rect.top)*o_scale_y/8);
	assert( point2 > 0 && point2 <= npoints );
	LineTo((proj_points[point2-1].h-proj_rect.left)*o_scale_x/8,
		   (proj_points[point2-1].v-proj_rect.top)*o_scale_y/8);
	
	while( (extra_pt_no = *args++ >> 16) != 0 ) //	while( (extra_pt_no = va_arg(args,long)) != 0 )
	{
//		message("extra_pt_no = %ld",extra_pt_no);
		assert( extra_pt_no > 0 && extra_pt_no <= npoints );
		LineTo((proj_points[extra_pt_no-1].h-proj_rect.left)*o_scale_x/8,
		   	   (proj_points[extra_pt_no-1].v-proj_rect.top)*o_scale_y/8);
	}
}

									// Select the color
void SpacePoints::color(const short color_no)
{
	assert( color_no < 16 );
	assert( thePort->picSave != (void *)0 );		// Picture should be under conctruction
	PicComment(color_no,0,(Handle)0);
}

/*
 *----------------------------------------------------------------------
 *				Pick up point/line coordinates from the window
 * 							(in dialog with the user)
 * 
 */

											// Pick up a point from the window
											// Remember to close any dialog (if any)
											// before calling this function
Point SpacePoints::pickup_point_from_window(void)
{
	const long event_timeout = 1;

	Rect port_rect = thePort->portRect;
									// The following constants are measured in 8th
									// of the original scale
	const int o_scale_x = (8*(port_rect.right-port_rect.left))/(proj_rect.right-proj_rect.left);
	const int o_scale_y = (8*(port_rect.top-port_rect.bottom))/(proj_rect.top-proj_rect.bottom);

	RgnHandle mouse_moving_rgn = NewRgn();	// It defines the region outside which mouse
									// moving should be considered an event
	
	RgnHandle window_rgn = window_inside_region();	// This is a region of our window in global coordinates
	RgnHandle screen_rgn = NewRgn();
	Point point;
	EventRecord event;
	WindowPtr wp;
	
	select();
	RectRgn(screen_rgn,&screenBits.bounds);
	CopyRgn(window_rgn,mouse_moving_rgn);
	SetCursor(*GetCursor(crossCursor));
	for(;;)
	{
		if( !WaitNextEvent(everyEvent, &event, event_timeout, mouse_moving_rgn) )
		  continue;

		switch(event.what)
		{
		  case activateEvt:
	    	if( is_our_window((WindowPtr)event.message) )
			  activate(event.modifiers & 0x01);
		     continue;

		  case updateEvt:
	    	if( is_our_window((WindowPtr)event.message) )
			  update();
		    continue;

		  case mouseDown:
		  point = event.where;
		  break;

		  case osEvt:
		  if( event.message & mouseMovedMessage == 0)
		    continue;
		  if( EqualRgn(window_rgn,mouse_moving_rgn) )
		  {
			XorRgn(mouse_moving_rgn,screen_rgn,mouse_moving_rgn);
		    SetCursor(&arrow);
		    continue;
		  }
		  else
		  {
			XorRgn(mouse_moving_rgn,screen_rgn,mouse_moving_rgn);
			SetCursor(*GetCursor(crossCursor));
		    continue;
		  }
		}
		break;
	}
	SetCursor(&arrow);
	GlobalToLocal(&point);
	DisposeRgn(mouse_moving_rgn);
	DisposeRgn(window_rgn);
	DisposeRgn(screen_rgn);

											// Convert from the visual plane
											// back to the user frame
	point.h = (point.h*8)/o_scale_x + proj_rect.left; 
	point.v = (point.v*8)/o_scale_y + proj_rect.top; 
	return point;
}


