/*
 ***********************************************************************
 *
 *					 	Stirring Spatial points
 *
 *	Defines the spatial points that can stir, i.e. move, rotate, etc.
 *
 ***********************************************************************
 */

#include "StirringPoints.h"
#include "Database.h"
#include <string.h>
#define _NOERRORCHECK_
#include <math.h>

/*
 *----------------------------------------------------------------------
 *						Constructing and Destructing
 */

StirringSpacialPoints::StirringSpacialPoints(void)
{
	rotation_axis_p1.x = 0;
	rotation_axis_p1.y = 0;
	rotation_axis_p1.z = 0;

	rotation_axis_p2.x = 0;
	rotation_axis_p2.y = 0;
	rotation_axis_p2.z = 1;
	
	rotation_angle = 0.314;
	setup_F();
}

StirringSpacialPoints::~StirringSpacialPoints(void)
{
}

void StirringSpacialPoints::set_rotation_point1(const int x, const int y, const int z)
{
	rotation_axis_p1.x = x;
	rotation_axis_p1.y = y;
	rotation_axis_p1.z = z;
}

/*
 *----------------------------------------------------------------------
 *					  Perform the rotation of the space point
 *
 * The following function performs rotation of a given point around the
 * axis by a given angle (in rads). The rotation is anti-clockwise for positive
 * values of angle and clockwise for the negative ones.
 *
 * To rotate a point P around an arbitrary spatial line (vector) (from point A1
 * to point A2), one needs
 *	- Change the reference frame such that new origin would coincide with a point
 *	  on the rotation axis, say, A1. It can be accomplished by translation the
 *	  point P by a vector -A1 = (-A1.x, -A1.y, -A1.z )'
 *  - Now, when the rotation axis goes through the origin, we can rotate the
 *    frame such that Oz would coincide with the rotation axis. This can be 
 *	  achieved by the rotation matrix Fy*Fx to be discussed later.
 *  - Rotation around axis Oz (i.e. on the xOy plane) within the new coordinate
 *    frame is extremely easy and can be done with the following matrix R
 *				| cos(th) -sin(th) 0 |
 *				| sin(th)  cos(th) 0 |   
 *				|   0		  0	   1 |
 *	- After the rotation is made, the coordinates should be transformed back
 *	  to the original coordinate system, performing rotation by F^(-1) = F and
 *	  translating back by the vector A1
 *
 *	  So, the resulting transformation can be written as
 *		P_transf = A1 + FxFyRFyFx(P-A1)
 *	  Note, that A1 and the matrix F depend only on the the direction of the axis and
 *	  do not depend on the point P and the rotation angle. Matrix R depends only
 *	  on the rotation angle.
 */

											// Print a fixed vector
static void print_vector(const char * title, float * fv)
{
	message("%s (%ld,%ld,%ld)/64 ",title,(long)(fv[0]*64),(long)(fv[1]*64),
			(int)(fv[2]*64));
}

									// Round the float gracefully
static long round(float fl)
{
	if( fl > 0 )
	  return (int)(fl + 0.5);
	else
	  return -(int)(-fl + 0.5);
}

SpacePoint StirringSpacialPoints::rotate(const SpacePoint pt, const float angle)
{
	float sinth = sin((double)angle);
	float costh = cos(angle);
	SpacePoint pt1;
	float w[3], w1[3];

//	message("pt: %ld, %ld, %ld ",pt.x,pt.y,pt.z);
	w[0] = pt.x - rotation_axis_p1.x;			// Perform translation
	w[1] = pt.y - rotation_axis_p1.y;
	w[2] = pt.z - rotation_axis_p1.z;
	
//	print_vector("w ",w);
													// Multiply by Fx
	w1[0] = w[0];
	w1[1] = cx*w[1] - sx*w[2];
	w1[2] = sx*w[1] + cx*w[2];
													// Multiply by Fy
	w[0] = cy*w1[0] - sy*w1[2];
	w[1] = w1[1];
	w[2] = sy*w1[0] + cy*w1[2];

//	print_vector("After F ",w);

													// Multiply by R
	w1[0]  = costh*w[0] - sinth*w[1];
	w1[1]  = sinth*w[0] + costh*w[1];
	w1[2]  = w[2];

													// Multiply by Fy transposed
	w[0] = cy*w1[0] + sy*w1[2];
	w[1] = w1[1];
	w[2] = -sy*w1[0] + cy*w1[2];
		
													// Multiply by Fx transposed
	w1[0] = w[0];
	w1[1] = cx*w[1] + sx*w[2];
	w1[2] = -sx*w[1] + cx*w[2];

//	print_vector("Final displacement ",w1);

	pt1.x = rotation_axis_p1.x + round(w1[0]);	// Perform translation back
	pt1.y = rotation_axis_p1.y + round(w1[1]);
	pt1.z = rotation_axis_p1.z + round(w1[2]);
	
//	_error("pt1.x %ld, pt1.y %ld, pt1.z %ld ",pt1.x,pt1.y,pt1.z);
	return pt1;
}
 
/*
 *----------------------------------------------------------------------
 *					Computing the frame rotation matrix F
 *
 * The present function computes the frame rotation matrix F such that
 * the line connecting the origin with the point (ax,ay,az) = 
 * (A2.x-A1.x,A2.y-A1.y,A2.z-A1.z) would become Oz in the new coordinates.
 * We can get it by performing two rotations, one around Ox axis to bring the line
 * into the plane zOx, and the other rotation around axis Oy to move the line on 
 * zOx to coincide with Oz.
 * The matrix of the first rotation, Fx, is to satisfy the following equation
 *
 *				| 1   0   0 | | ax |   | ax  |
 *				| 0  cx -sx | | ay | = | 0   |
 *				| 0  sx  cx | | az |   | az1 |
 *
 *	where cx^2 + sx^2 = 1
 *
 *  It all comes down to cx*ay - sx*az = 0, that along with the eqaution just
 *  above gives us
 *			cx = az/an1,  sx = ay/an1, an1 = sqrt(ay^2+az^2)
 *  If both ay and az happen to be zero, the line is already within the plane zOx
 *	(as a matter of fact, it coincides with Ox), and we just can set
 *			cx = 1 and sx = 0
 *
 *  The matrix of the second rotation needs to satisfy
 *
 *				| cy  0  -sy | | ax  |   | 0   |
 *				|  0  1    0 | | 0   | = | 0   |
 *				| sy  0  cy  | | az1 |   | az2 |
 *
 *	where cy^2 + sy^2 = 1 and az1 = ay*sx + az*cx
 *
 *	Similarly to the previous example, we obtain
 *			cy = az1/an2, sy = ax/an2, an2 = sqrt(ax^2 + az1^2)
 *  Note, unless the vector (ax,ay,az) is zero (which it is not), an2 is never 0.
 *
 *  To cut down on spice and time, matrix Fx is represented only by cx and sx, and
 *  similarly is matrix Fy.
 *
 */

void StirringSpacialPoints::setup_F(void)
{
	short ax = rotation_axis_p2.x - rotation_axis_p1.x;
	short ay = rotation_axis_p2.y - rotation_axis_p1.y;
	short az = rotation_axis_p2.z - rotation_axis_p1.z;
	float az1;
	long an1_sqr = ay*ay + az*az;
	float an2;
	
	if( an1_sqr == 0 )
	  cx = 1.0, sx = 0;
	else
	  cx = az/sqrt((double)an1_sqr),
	  sx = ay/sqrt((double)an1_sqr);

	  
	az1 = ay*sx + az*cx;
	
	an2 = sqrt((double)(ax*ax + az1*az1));
	cy = az1/an2;
	sy = ax/an2; 
}



/*
 *----------------------------------------------------------------------
 *					Define the reply to the mouse click
 */

void StirringSpacialPoints::click(void)
{
	register int i;
#if 0
	SpacePoint p; p.x=-1; p.y=1; p.z=4; rotate(p,atan(1.0)*2);
#endif

	if( TD_rotate_with_mouse )				// Reset the 'rotate with mouse' mode
	{
	   TD_rotate_with_mouse = FALSE;
	   return;
	}
	
	for(i=0; i<npoints; i++)				// Perform the rotation
	  points[i] = rotate(points[i],rotation_angle);
	  
	refresh();
}

/*
 *----------------------------------------------------------------------
 *					Define the reply to the "time click"
 * Check if the mouse has moved, and if has, perform the rotation around
 * the axis perpendicular to the mouse movement and by angle depending on
 * the direction mouse moved.
 *
 */

										// return the point that is closer to p2
										// than p1 yet lies on p2->p1
static SpacePoint not_so_far(SpacePoint p1,SpacePoint p2)
{
	SpacePoint displ;
	SpacePoint new_p1;
	long dist_sqr;
	long dist;
	displ.x = p1.x - p2.x;
	displ.y = p1.y - p2.y;
	displ.z = p1.z - p2.z;
	dist_sqr = displ.x*displ.x + displ.y*displ.y + displ.z*displ.z;

	if( dist_sqr == 0 )
	{
	   new_p1 = p1;
	   new_p1.x = p1.x + 1;
	   return new_p1;
	}

#if 1
	if( dist_sqr < 64 )
	  return p1;
	
	dist = 2 + dist_sqr/8;			// This is a Newton approximation to a square root
	new_p1.x = p2.x + 64*displ.x/dist;
	new_p1.y = p2.y + 64*displ.y/dist;
	new_p1.z = p2.z + 64*displ.z/dist;
#endif
	return new_p1;
}

struct MouseReply
{
	SpacePoint p;
	long distance;
	short direction;					// +1 if mouse moved up and right, -1 otherwise
};

    									// Compute the rotation axis to go through
    									// p1 and perpendicular to the
    									// line now_mouse - prev_mouse
    									// p2 = a1 + (a2-a1)
    									//  = (a2-a1)(p1-a1)/|a2-a1|^2
    									// where a1 = prev_mouse, a2 = now_mouse
    									// The function returns p2 and the distance
    									// |a2-a1|^2 the mouse has moved at
static struct MouseReply account_for_mouse_move(const SpacePoint p1)
{
	struct MouseReply mr;
  	static Point prev_mouse;
    Point now_mouse;
    SpacePoint mouse_moved;
    SpacePoint p1a1;							// Vector p1 - a1
	SpacePoint p2p1;							// Vector p2 - p1
    float mu;
    
    GetMouse(&now_mouse);
    
    mouse_moved.x = 0;							// Compute the distance the mouse moved
    mouse_moved.y = now_mouse.h - prev_mouse.h;
    mouse_moved.z = now_mouse.v - prev_mouse.v;
    mr.distance   = mouse_moved.x*mouse_moved.x +
    				     mouse_moved.y*mouse_moved.y +
    				     mouse_moved.z*mouse_moved.z;
    if( mr.distance == 0 )				// The mouse hasn't been moved at all
      return mr;
   
   	if( mouse_moved.y >= 0 && mouse_moved.z >= 0 )
   	  mr.direction = +1;
   	else
   	  mr.direction = -1;
   	   
    p1a1.x = p1.x;							// Compute |p1-a1|
    p1a1.y = p1.y - prev_mouse.h;
    p1a1.z = p1.z - prev_mouse.v;
    mu = (p1a1.x*mouse_moved.x + p1a1.y*mouse_moved.y + p1a1.z*mouse_moved.z)/
       (double)mr.distance;

    mr.p.x = 0 + round(mu*mouse_moved.x);
    mr.p.y = prev_mouse.h + round(mu*mouse_moved.y);
    mr.p.z = prev_mouse.v + round(mu*mouse_moved.z);

	mr.p = not_so_far(mr.p,p1);

    prev_mouse = now_mouse;
	return mr;
}

void StirringSpacialPoints::timeclick(void)
{
  static long int prev_tick_count;
  
  if( TD_rotate_with_mouse )
  {										// Perform the rotation as mouse moves
    register int i;
	float angle;
	struct MouseReply mr = account_for_mouse_move(rotation_axis_p1);

	if( mr.distance == 0 )
	  return;									// The mouse has been moved
	
	rotation_axis_p2 = mr.p;
	
	if( rotation_axis_p2.x == rotation_axis_p1.x &&
	    rotation_axis_p2.y == rotation_axis_p1.y &&
	    rotation_axis_p2.z == rotation_axis_p1.z )
	  _error("zero distance");
	  
    setup_F();									// Setup the rotation axis
    
    angle = rotation_angle * mr.direction;

//	message("angle %ld/100",(int)(100*angle)); 
    
	for(i=0; i<npoints; i++)
	  points[i] = rotate(points[i],rotation_angle);
	  
	refresh();
    
  }
  
  if( !TD_animation )
    return;
    
  if( TickCount() - prev_tick_count > TD_animation_period )
    prev_tick_count = TickCount(),
    click();
}

